#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <unistd.h>
#include <wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>

extern int errno;

/* alza il semaforo */
int up_sem(int semid, int semnum) {
  struct sembuf sops = { semnum, -1, SEM_UNDO };
  
  if (semop(semid, &sops, 1)) {
    perror("up_sem() - Error up semaphore");
    return -1;
  } 
  return 0;
}

/* attende che il semaforo sia 0 */
int wait_sem(int semid, int semnum) {
  struct sembuf sops = { semnum, 0, SEM_UNDO };
  
  if (semop(semid, &sops, 1)) {
    perror("wait_sem() - Error wait semaphore");
    return -1;
  } 
  return 0;
}

/* stampa l'uso */
void usage(const char *me) {
    fprintf(stdout, "Usage: %s <process number>\n", me);
}

/* stampa un po' di informazioni sull'esecuzione :
 *  - sleep
 *  - arrivo alla barriera
 *  - ripresa esecuzione
 */
void go_to_the_barrier(int semid, unsigned short n) {
  struct timeval time_now;
  unsigned long sleeptime;
  
  if (gettimeofday(&time_now, NULL)!=0) {
    fprintf(stderr, "<%d>go_to_the_barrier(): %s\n", n, strerror(errno));
    return;
  }
  sleeptime = 1 + (time_now.tv_sec % n );
  fprintf(stdout, "%3d%s child: sleeping for %ld sec\n", n, 
                                                       n==1 ? "st" : (
                                                       n==2 ? "nd" : (
                                                       n==3 ? "rd" : "th")),
                                                       sleeptime);
  sleep(sleeptime);

  fprintf(stdout, "%3d%s child: getting to the barrier\n", n,
                                                         n==1 ? "st" : (
                                                         n==2 ? "nd" : (
                                                         n==3 ? "rd" : "th")));
  /* barriera */
  up_sem(semid, 0);
  
  if (wait_sem(semid, 0))
    fprintf(stderr, "<%d>go_to_the_barrier(): wait_sem() returned -1\n", n);
  
  fprintf(stdout, "%3d%s child: resuming\n", n,
                                            n==1 ? "st" : (
                                            n==2 ? "nd" : (
                                            n==3 ? "rd" : "th")));
  
  return;
}

/*
 *  MAIN
 */
int main (int argc, char *argv[]) {

  int semid, status;
  unsigned short pnum, count;

  /* controllo argomenti */
  if (argc < 2) {
    usage(argv[0]);
    return 0;
  }
  if ((pnum = atoi(argv[1])) < 1) {
    fprintf(stderr, "Error: <process number> must be >= 1\n");
    usage(argv[0]);
    return 0;
  }

  /* creazione set di semafori */
  semid = semget(IPC_PRIVATE, 1, IPC_CREAT | IPC_EXCL | 0600); 
  if (semid < 0) {
    fprintf(stderr, "main() - Error creating semaphore set: %s\n", strerror(errno));
    return 1;
  }
  /* inizializzazione set di semafori */
  if (semctl (semid, -1, SETALL, &pnum) < 0) {
    fprintf(stderr, "main() - Error intializing semaphore set: %s\n", strerror(errno));
    goto out;
    return 1;
  }

  /* fork n times */
  for (count=1 ; count<=pnum; count++) {
    switch(fork()) {
      case 0:
        /* figlio: va alla barriera */
        go_to_the_barrier(semid, count);
        return 0;

      case -1:
        fprintf(stderr, "main() - Error forking %d%s child: %s\n", count, 
                                                                   count==1 ? "st" : (
                                                                   count==2 ? "nd" : (
                                                                   count==3 ? "rd" : "th")),
                                                                   strerror(errno));
    }
  }
  
  /* aspetta l'uscita di tutti i figli */
  for (count=1 ; count<=pnum; count++)
    wait(&status);
  
out:
  if (semctl(semid, -1, IPC_RMID) < 0) {
    fprintf(stderr, "main() - Error removing semaphore set: %s\n", strerror(errno));
    return 1;
  }
  return 0;
}
